/* recover.cpp -- implements the method described in my paper "A method for
 *                recovering arbitrary graphs", 1/4/05.
 *
 * Copyright (C) 2003-2005 Nicolas Addington (adding@math.wisc.edu)
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version. */


/* INPUT FORMAT
 *
 * An input file gives the total number of nodes, the number of boundary nodes,
 * and the signs of the above-diagonal entries of the Kirchhoff matrix (a minus
 * if there is an edge and a zero if there is none), all separated by
 * whitespace.  For example,
 *   6 4
 *      0 0 - 0 -
 *        0 0 0 -
 *          0 - 0
 *            - 0
 *              -
 * has six nodes, four of which are boundary nodes (hence two interior nodes),
 * and there are edges between nodes 1 and 4, 1 and 6, 2 and 6, 3 and 5, 4 and
 * 5, and 5 and 6.  The entries do not have to be aligned as they are in this
 * example, nor do there have to be line breaks. */


#include <iostream>
#include <queue>

using namespace std;

// some function prototypes
char *matrix_name(int, int);
void cascade(int, int, int, int);
void if_got_one_get_the_other(int, int, int, int, int, int);
int det(int, int, bool*, bool*, int);
int det_internal(int, bool*, bool*, int, int, int, int, int);

/* alloc_ary(type, lo, hi): allocate array of type 'type' with indices 'lo' to
 * 'hi'.  sorry to have so few zero-based arrays, but otherwise my head would
 * explode going back and forth between the code and the paper. */
#define alloc_ary(type, lo, hi) (new type[(hi)-(lo)+1] - (lo))

// free_ary(ary, lo): free array 'ary' with lower bound 'lo'
#define free_ary(ary, lo) delete[] ((ary) + (lo))

int nn; // total nodes
int bd; // boundary nodes

// graphs[n][i][j] is the sign of the i, j entry of K_n
int ***graphs;

/* R[lo][hi] is the residue matrix from the elimination of nodes lo to hi, but
 * it is also convenient to let R[n+1][nn+1] be K[n].  R[lo][hi][i][j] = t if
 * the i, j entry of R[lo][hi] was recovered at time t, or the maximum possible
 * unsigned int "infty" if it has yet to be recovered.  all zero entries are
 * recovered at the same time, 0.  this may seem unnecessarily complicated; see
 * the comment at queue_item::process for why a boolean value is not enough. */
unsigned ****R;
unsigned now = 0; // will be incremented each time we recover something
#define infty (~0)

// is_zero[lo][hi][i] is true iff the i^th row/column of R[lo][hi] is all zeros.
bool ***is_zero;

// unknown_entries[lo][hi] = number of unknown entries in R[lo][hi]
int **unknown_entries;

// submatrices that may yield new entries go in this queue
class queue_item {
public:
  int lo, hi;
  bool *rows, *cols;
  int size;

  int unknown_row, unknown_col;

  unsigned dob;
  int new_corner_row, new_corner_col;

  queue_item(int, int, int, int);
  queue_item(queue_item*, int, int, int, int);
  ~queue_item();

  int singular_size();

  void process();
};

struct compare_qi { bool operator()(queue_item*, queue_item*); };
priority_queue <queue_item*, vector <queue_item*>, compare_qi> q;


int main(int argc, char **argv) {
  // we should check the validity of the input, but we don't.

  // read nn (total nodes), bd (boundary nodes)
  cin >> nn >> bd;

  // allocate graphs
  graphs = alloc_ary(int**, bd, nn);
  for (int n = bd; n <= nn; n++) {
    graphs[n] = alloc_ary(int*, 1, n);
    for (int i = 1; i <= n; i++)
      graphs[n][i] = alloc_ary(int, 1, n);
  }

  // read the graph
  for (int i = 1; i <= nn; i++) {
    graphs[nn][i][i] = 1;
    for (int j = i+1; j <= nn; j++) {
      char c; cin >> c;
      graphs[nn][i][j] = graphs[nn][j][i] = c == '0' ? 0 : -1;
    }
  }

  // star-k each interior node (initialize remaining graphs)
  for (int n = nn - 1; n >= bd; n--) {
    for (int i = 1; i <= n; i++) {
      graphs[n][i][i] = 1;
      for (int j = i+1; j <= n; j++)
        graphs[n][i][j] = graphs[n][j][i] = graphs[n+1][i][j] ||
          (graphs[n+1][i][n+1] && graphs[n+1][j][n+1]) ? -1 : 0;
    }
  }

  // allocate and initialize R
  R = alloc_ary(unsigned***, bd+1, nn+1);
  for (int lo = bd+1; lo <= nn+1; lo++) {
    R[lo] = alloc_ary(unsigned**, lo, nn+1);
    for (int hi = lo; hi <= nn+1; hi++) {
      R[lo][hi] = alloc_ary(unsigned*, 1, lo-1);
      for (int i = 1; i <= lo-1; i++) {
        R[lo][hi][i] = alloc_ary(unsigned, 1, lo-1);
        for (int j = 1; j <= lo-1; j++)
          R[lo][hi][i][j] = infty;
      }
    }
  }

  // allocate and initialize is_zero
  is_zero = alloc_ary(bool**, bd+1, nn+1);
  for (int lo = nn+1; lo >= bd+1; lo--) {
    is_zero[lo] = alloc_ary(bool*, lo, nn+1);
    for (int hi = lo; hi <= nn+1; hi++) {
      is_zero[lo][hi] = alloc_ary(bool, 1, lo-1);
      if (hi == nn+1) // Kirchhoff matrix
        for (int i = 1; i <= lo-1; i++)
          is_zero[lo][hi][i] = false;
      else if (lo == hi) // one-layer residue matrix
        for (int i = 1; i <= lo-1; i++)
          is_zero[lo][hi][i] = graphs[lo][i][lo] == 0;
      else // multi-layer residue matrix
        for (int i = 1; i <= lo-1; i++)
          is_zero[lo][hi][i] = is_zero[lo][lo][i] && is_zero[lo+1][hi][i];
    }
  }

  // allocate and initialize unknown_entries
  unknown_entries = alloc_ary(int*, bd+1, nn+1);
  for (int lo = bd+1; lo <= nn+1; lo++) {
    unknown_entries[lo] = alloc_ary(int, lo, nn+1);
    for (int hi = lo; hi <= nn+1; hi++)
      unknown_entries[lo][hi] = (lo-1)*lo/2;
  }

  // start adding zeros

  // we know all the zeros in the one-layer residue matrices
  for (int lo = bd+1; lo <= nn; lo++)
    for (int i = 1; i <= lo-1; i++)
      for (int j = i; j <= lo-1; j++)
        if (is_zero[lo][lo][i] || is_zero[lo][lo][j])
          cascade(lo, lo, i, j);

  // we know all the zeros in the Kirchhoff matrix
  for (int i = 1; i <= nn; i++)
    for (int j = i+1; j <= nn; j++)
      if (graphs[nn][i][j] == 0)
        cascade(nn+1, nn+1, i, j);

  // start adding non-zero entries
  now = 1;

  // we know the whole response matrix
  for (int i = 1; i <= bd; i++)
    for (int j = i; j <= bd; j++)
      cascade(bd+1, nn+1, i, j);

  // if requested, print what we know at this point
  if (argc >= 2 && strcmp(argv[1], "--showme") == 0) {
    for (int lo = bd+1; lo <= nn+1; lo++)
      for (int hi = lo; hi <= nn+1; hi++) {
        cout << matrix_name(lo, hi) << endl;
        for (int i = 1; i <= lo-1; i++) {
          for (int j = 1; j <= lo-1; j++)
            switch (R[lo][hi][i][j]) {
              case 0 : cout << " 0"; break;
              case infty : cout << "  "; break;
              default : cout << " x"; break;
            }
          cout << endl;
        }
        cout << endl;
      }
  }

  // keep track of how many parameters we introduce
  int parameters = 0;

  // process queue
  while (bd < nn) {
    // if we know a whole Kirchhoff matrix, stop looking above it
    while (bd < nn && unknown_entries[bd+2][nn+1] == 0) bd++;

    if (q.empty()) {
      // parametrize
      for (int i = 1; i <= bd+1; i++)
        for (int j = i+1; j <= bd+1; j++)
          if (R[bd+2][nn+1][i][j] == infty) {
            parameters++;
            cascade(bd+2, nn+1, i, j);
            cout << matrix_name(bd+2, nn+1) << ": parametrized " << i << "," << j << endl;
            i = j = bd+2; // exit 2 for loops
          }
    } else {
      queue_item *qi = q.top();
      q.pop();
      if (qi->lo >= bd+1 && unknown_entries[qi->lo][qi->hi])
        qi->process();
      delete qi;
    }
  };

  // print result
  cout << endl;
  switch (parameters) {
    case 0 :
      cout << "Graph is recoverable.";
      break;
    case 1 :
      cout << "Graph is recoverable with 1 parameter.";
      break;
    default :
      cout << "Graph is recoverable with " << parameters << " parameters.";
  }
  cout << endl;

  // let the garbage collect itself when the process terminates

  return 0;
}

/* matrix_name
 *
 * return "R^lo_hi" or "K_{lo-1}", as appropriate. */
char *matrix_name(int lo, int hi) {
  static char *ret = new char[16];
  if (hi <= nn) // residue matrix
    sprintf(ret, "R^%d_%d", lo, hi);
  else // Kirchhoff matrix
    sprintf(ret, "K_%d", lo-1);
  return ret;
}


/* cascade(lo, hi, i, j)
 *
 * - mark R[lo][hi][i][j] as recovered
 * - add to queue for determinant checking later
 * - get the entries in other matrices that can be gotten for cheap,
 *   i.e. without determinants */
void cascade(int lo, int hi, int i, int j) {
  // if already known
  if (R[lo][hi][i][j] < infty) return;

  // mark R[lo][hi][i][j] as recovered
  R[lo][hi][i][j] = R[lo][hi][j][i] = now;
  unknown_entries[lo][hi]--;

  // if we're getting non-zero entries
  if (now > 0) {
    // increment time
    now++;
    // add to queue for determinant checking later
    q.push(new queue_item(lo, hi, i, j));
  }

  /* VERTICAL STUFF: Suppose we're working with
   *
   *   5 56 567 5678 56789
   *      6  67  678  6789
   *          7   78   789
   *               8    89
   *                     9 .
   *
   * We have equations of the form 56 + 789 = 56789, so when we have two terms,
   * we can get the third. */

  // to the east: suppose we were given the first term 56
  for (int k = hi+1; k <= nn+1; k++)
    if_got_one_get_the_other(lo,k, hi+1,k, i,j);

  // to the north: suppose we were given the second term 789
  int max = bd+1; // to avoid looking at e.g. the 6,6 entry of a 5x5 matrix
  if (i+1 > max) max = i+1;
  if (j+1 > max) max = j+1;
  for (int k = max; k < lo; k++)
    if_got_one_get_the_other(k,lo-1, k,hi, i,j);

  // to the west and south: suppose we were given the third term 56789
  for (int k = lo; k < hi; k++)
    if_got_one_get_the_other(lo,k, k+1,hi, i,j);


  /* HORIZONTAL STUFF:  Rows of Kirchhoff matrices sum to zero.  The i^th row
   * of R^n_n sums to the i,n entry of K_n, but this might be redundant. */
  if (hi == nn+1 || (lo == hi && hi <= nn)) {
    int count_unknown_i = 0, // # of unknown entries in row/column i
        where_unknown_i,     // position of last such entry
        count_unknown_j = 0, // # of unknown entries in row/column j
        where_unknown_j;     // position of last such entry
    for (int k = 1; k <= lo-1; k++) {
      if (R[lo][hi][i][k] == infty) { count_unknown_i++; where_unknown_i = k; }
      if (R[lo][hi][j][k] == infty) { count_unknown_j++; where_unknown_j = k; }
    }

    if (hi == nn+1) { // Kirchhoff matrix
      if (count_unknown_i == 1) cascade(lo, hi, i, where_unknown_i);
      if (count_unknown_j == 1) cascade(lo, hi, j, where_unknown_j);
    } else if (lo == hi) { // one-layer residue matrix
      // is this necessary?
      if (count_unknown_i == 0) cascade(lo+1, nn+1, i, lo);
      if (count_unknown_j == 0) cascade(lo+1, nn+1, j, lo);
      if (count_unknown_i == 1 && R[lo+1][nn+1][i][lo] < infty)
        cascade(lo, hi, i, where_unknown_i);
      if (count_unknown_j == 1 && R[lo+1][nn+1][j][lo] < infty)
        cascade(lo, hi, j, where_unknown_j);
    }
  }
}


/* if_got_one_get_the_other
 *
 * if one of R[lo1][hi1][i][j] and R[lo2][hi2][i][j] is known, get the other.
 * used by cascade. */
void if_got_one_get_the_other(int lo1, int hi1, int lo2, int hi2, int i, int j) {
  if (R[lo1][hi1][i][j] < infty && R[lo2][hi2][i][j] == infty)
    cascade(lo2, hi2, i, j);
  else if (R[lo1][hi1][i][j] == infty && R[lo2][hi2][i][j] < infty)
    cascade(lo1, hi1, i, j);
}


/* class queue_item
 *
 * A queue item says we are missing zero or one entries in a square:
 *
 *   . . . . .        . . . . .
 *   . . x x x        . . x x x
 *   . . x x x   or   . . o x x
 *   . . . . .        . . . . .
 *   . . x x x        . . x x x  .
 *
 * In queue_item::process, we try to recover the missing entry, and if that
 * fails (or if there is none), we look for opposite corners
 *
 *   X . . . .        X . . . .
 *   . . x x x        . . x x x
 *   . . x x x   or   . . o x x
 *   . . . . .        . . . . .
 *   . . x x x        . . x x x
 *
 * such that if we add them to make a larger queue_item
 *
 *   X . x o x        X . x x x
 *   x . x x x        x . x x x
 *   x . x x x   or   x . o x x
 *   . . . . .        . . . . .
 *   x . x x x        x . x x x
 *
 * we're still missing at most one entry. */


/* queue_item constructor.
 * make a new square in R[lo][hi] consisting only of row i and column j */
queue_item::queue_item(int _lo, int _hi, int i, int j) {
  lo = _lo;
  hi = _hi;
  rows = alloc_ary(bool, 1, nn);
  cols = alloc_ary(bool, 1, nn);
  for (int k = 1; k <= lo-1; k++) {
    rows[k] = k == i;
    cols[k] = k == j;
  }
  size = 1;

  // nothing missing
  unknown_row = unknown_col = 0;

  // redundancy avoidance mechanism (see queue_item::procecss)
  dob = R[lo][hi][i][j];
  new_corner_row = lo;
  new_corner_col = lo;
}


/* queue_item constructor.
 * copy qi and add row i and column j */
queue_item::queue_item(queue_item *qi, int i, int j, int _unknown_row, int _unknown_col) {
  lo = qi->lo;
  hi = qi->hi;
  rows = alloc_ary(bool, 1, nn);
  cols = alloc_ary(bool, 1, nn);
  for (int k = 1; k <= lo-1; k++) {
    rows[k] = qi->rows[k] || k == i;
    cols[k] = qi->cols[k] || k == j;
  }
  size = qi->size + 1;

  unknown_row = _unknown_row;
  unknown_col = _unknown_col;

  // redundancy avoidance mechanism (see queue_item::process)
  dob = qi->dob;
  new_corner_row = i;
  new_corner_col = j;
}


// queue_item destructor
queue_item::~queue_item()
  { free_ary(rows, 1); free_ary(cols, 1); }


/* queue_item::singular_size
 *
 * in R[lo][hi], submatrices of this size or larger must be singular.
 * e.g. a 3x3 submatrix of a 2-layer residue matrix must be singular */
int queue_item::singular_size() {
  if (hi <= nn) // residue matrix
    return hi-lo+1 + 1;
  else // Kirchhoff matrix
    return lo-1;
}


/* queue_item::process
 *
 * See comment above at 'class queue_item' for what this does.
 *
 * We pretend we only know about entries that were known before the queue item
 * was created and insist that our new corners are to the northwest so that the
 * same squares don't get checked over and over. */
void queue_item::process() {
  // if we're missing an entry, try to recover
  if (unknown_row) {
    // if it's been found in the interrim, we're done
    if (R[lo][hi][unknown_row][unknown_col] < infty) return;

    // calculate the small determinant
    rows[unknown_row] = cols[unknown_col] = false;
    int small_det = det(lo, hi, rows, cols, size-1);
    rows[unknown_row] = cols[unknown_col] = true;

    // if small matrix is invertible
    if (small_det == 1 || small_det == -1)
      // and large matrix is singular
      if (size == singular_size() || det(lo, hi, rows, cols, size) == 0) {
        // then recover
        cascade(lo, hi, unknown_row, unknown_col);
        // and say we did
        cout << matrix_name(lo, hi) << ": recovered " << unknown_row << "," << unknown_col << " from";
        for (int k = 1; k <= lo-1; k++) if (rows[k]) cout << " " << k;
        cout << " x";
        for (int k = 1; k <= lo-1; k++) if (cols[k]) cout << " " << k;
        cout << endl;
        return;
      }

    // if there's nothing larger, quit
    if (size == lo-1) return;
  }

  // look for new corners to add

  // i^th entry is # of unknown entries above/below our square in row i
  static int *count_unknown_in_row = alloc_ary(int, 1, nn);
  // positions of last unknown entry in each row
  static int *where_unknown_in_row = alloc_ary(int, 1, nn);
  // j^th entry is # of unknown entries left/right of our square in column j
  static int *count_unknown_in_col = alloc_ary(int, 1, nn);
  // positions of last unknown entry in each column
  static int *where_unknown_in_col = alloc_ary(int, 1, nn);

  for (int i = 1; i <= lo-1; i++) if (!rows[i] && !is_zero[lo][hi][i]) {
    count_unknown_in_row[i] = 0;
    for (int j = 1; j <= lo-1; j++)
      if (cols[j] && R[lo][hi][i][j] >= dob)
        { count_unknown_in_row[i]++; where_unknown_in_row[i] = j; }
  }

  for (int j = 1; j <= lo-1; j++) if (!cols[j] && !is_zero[lo][hi][j]) {
    count_unknown_in_col[j] = 0;
    for (int i = 1; i <= lo-1; i++)
      if (rows[i] && R[lo][hi][i][j] >= dob)
        { count_unknown_in_col[j]++; where_unknown_in_col[j] = i; }
  }

  // if the square root trick is possible, remember where
  int sqrt_trick = 0;
  if (hi <= nn && lo == hi && size == 1) {
    int i = 1; while (!rows[i]) i++;
    if (cols[i]) sqrt_trick = i;
  }

  // check opposite corners
  for (int i = 1; i < new_corner_row; i++) if (!rows[i] && !is_zero[lo][hi][i])
    for (int j = 1; j < new_corner_col; j++) if (!cols[j] && !is_zero[lo][hi][j])
      // if we add i,j as our corner, how many entries will be missing?
      switch ((unknown_row != 0) + (R[lo][hi][i][j] >= dob) + count_unknown_in_row[i] + count_unknown_in_col[j]) {
        case 0 :
          // if the new guy is necessarily singular, it can't help
          if (size+1 < singular_size())
            q.push(new queue_item(this, i, j, unknown_row, unknown_col));
          break;
        case 1 :
          // if we're necessarily singular, anything larger can't help
          if (size < singular_size()) {
            if (unknown_row)
              q.push(new queue_item(this, i, j, unknown_row, unknown_col));
            else if (R[lo][hi][i][j] >= dob)
              q.push(new queue_item(this, i, j, i, j));
            else if (count_unknown_in_row[i])
              q.push(new queue_item(this, i, j, i, where_unknown_in_row[i]));
            else if (count_unknown_in_col[j])
              q.push(new queue_item(this, i, j, where_unknown_in_col[j], j));
          }
          break;
        case 2 : // square root trick
          if (sqrt_trick && i == j && R[lo][hi][i][sqrt_trick] == infty) {
            cascade(lo, hi, i, sqrt_trick);
            cout << matrix_name(lo, hi) << ": recovered " << i << "," << sqrt_trick << " using square root trick" << endl;
          }
          break;
      }
}


/* compare_qi::operator()
 *
 * returns true if qi1 is to come off the priority_queue *after* qi2 (don't
 * blame me, blame the STL).  prefer queue_items with
 *  - smaller 'lo' index (closer to the action), or in case of a tie
 *  - smaller 'hi' index (thinner residue matrix), or in case of a tie
 *  - smaller size (easier to take determinants), or in case of a tie
 *  - higher dob (new queue_items can see old entries, but not vice versa)
 *
 * if you want to make this program run faster, start here - see section 5,
 * question 7 in my paper */
bool compare_qi::operator()(queue_item *qi1, queue_item *qi2) {
  if (qi1->lo != qi2->lo)
    return qi1->lo > qi2->lo;
  else if (qi1->hi != qi2->hi)
    return qi1->hi > qi2->hi;
  else if (qi1->size != qi2->size)
    return qi1->size > qi2->size;
  else
    return qi1->dob < qi2->dob;
}


/* det
 *
 * returns 1, 0, or -1 if that's the sign of the determinant, or 2 if we can't
 * decide from the signs of K alone.
 *
 * it would be great to cache the result, but I can't think can't think of an
 * efficient way to do it. */
int det(int lo, int hi, bool *rows, bool *cols, int size) {
  if (size == 1) { // shortcut for 1x1's
    int i = 1; while (!rows[i]) i++;
    int j = 1; while (!cols[j]) j++;
    if (hi <= nn) // residue matrix
      return R[lo][hi][i][j] == 0 ? 0 : -1;
    else // Kirchhoff matrix
      return graphs[lo-1][i][j];
  }

  if (hi <= nn) { // residue matrix
    // tack on extra rows and columns
    for (int i = lo; i <= hi; i++)
      rows[i] = cols[i] = true;
    return det_internal(hi, rows, cols, lo-1, size, hi-lo+1, size, hi-lo+1);
  } else // Kirchhoff matrix
    return det_internal(lo-1, rows, cols, 0, 0, size, 0, size);
}

int det_internal(int n,        // we're talking about graphs[n]
                 bool *rows, bool *cols,
                 int suppress, /* upper left this-by-this corner
                                * is changed to zeros (suppressed) */
                 int rows_z,   // number of suppressed rows
                 int rows_nz,  // number of unsuppressed rows
                 int cols_z,   // number of suppressed columns
                 int cols_nz   // number of unsuppressed columns
                ) {
  // if too much is suppressed, it's singular
  if (rows_z > rows_nz && cols_z > cols_nz) return 0;

  // shortcut for 1x1 determinants
  // case where it's in the suppressed corner is covered above
  if (rows_z + rows_nz == 1) {
    int i = 1; while (!rows[i]) i++;
    int j = 1; while (!cols[j]) j++;
    return graphs[n][i][j];
  }

  // count non-zero entries in each row, column
  static int *row_counts = alloc_ary(int, 0, nn),
             *col_counts = alloc_ary(int, 0, nn);
  row_counts[0] = col_counts[0] = n+1;
  for (int i = 1; i <= n; i++)
    row_counts[i] = row_counts[i] = 0;
  for (int i = 1; i <= n; i++) if (rows[i])
    for (int j = 1; j <= n; j++) if (cols[j])
      if ((i > suppress || j > suppress) && graphs[n][i][j])
        { row_counts[i]++; col_counts[j]++; }

  // find row, column with fewest
  int row_sign = 1, short_row = 0, short_row_sign,
      col_sign = 1, short_col = 0, short_col_sign;
  for (int i = 1; i <= n; i++) {
    if (rows[i]) {
      // if any row is all zeros, shortcut
      if (row_counts[i] == 0) return 0;
      row_sign *= -1;
      if (row_counts[i] < col_counts[short_row]) {
        short_row = i;
        short_row_sign = row_sign;
      }
    }
    if (cols[i]) {
      // if any column is all zeros, shortcut
      if (col_counts[i] == 0) return 0;
      col_sign *= -1;
      if (col_counts[i] < col_counts[short_col]) {
        short_col = i;
        short_col_sign = col_sign;
      }
    }
  }

  // if some column has more zeros than the best row, transpose
  bool transpose = col_counts[short_col] < row_counts[short_row];
  if (transpose) {
    swap(rows, cols);
    short_row = short_col;
    short_row_sign = short_col_sign;
  }

  // expand along row with most zeros
  int ret = 0, i = short_row, sign = short_row_sign;
  for (int j = 1; j <= n; j++) if (cols[j]) {
    sign *= -1;
    int a = i <= suppress && j <= suppress ? 0 : graphs[n][i][j];
    if (a == 0) continue;
    rows[i] = cols[j] = false;
    int A = det_internal(n, rows, cols, suppress,
                         rows_z - (i <= suppress),
                         rows_nz - (i > suppress),
                         cols_z - (j <= suppress),
                         cols_nz - (j > suppress)
                        );
    rows[i] = cols[j] = true;
    if (A == 0) continue;
    if (A == 2) { ret = 2; break; }
    if (ret == 0) { ret = a*A*sign; continue; }
    if (ret != a*A*sign) { ret = 2; break; }
  }

  // if we transposed, transpose back
  if (transpose)
    swap(rows, cols);

  return ret;
}
